% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008-2013, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.

:- module(tokenizer, [
		tokenize/2,
    expand_contracted_forms/3
	]).
:- use_module(library(error)).

:- use_module('../lexicon/chars', [
		is_letter/1,
		is_digit/1
	]).


:- use_module('../logger/error_logger').

:- use_module('../utils/ace_niceace', [
        pronoun_split/2
    ]).


/** <module> APE Tokenizer

@author Kaarel Kaljurand
@author Tobias Kuhn
@version 2010-03-28

Comments:

* Strings (between double quotes) are tokenized into [", content, "].
This can be misleading, e.g. in the sentence: `"1" represents 1.'
the verb `represents' is reported as the 4th token.
One could instead produce the term string(content), this would
probably also fix the buggy handling of `Every dot is ".".'.

* Saxon Genitives are tokenized as [noun, ', s] and [nouns, ']

* Digits cannot start a token that contains other symbols than digits and dots.
Consider the following input string and its tokenization.

==
123man man123 123man123 man123man

[123, man, man123, 123, man123, man123man]
==

BUG: future work: add character counting to be able to report character offsets
together with every token.

Example:

?- tokenizer:codes_to_tokens("2men;can't like\"#@\"and:everything@.", T), writeq(T).
[2, men, can, not, like, '"#@"', and, :, every, '-thing', '.']

*/

:- use_module('../utils/logicmoo_ape_utils').

:- if( \+ current_predicate(is_codelist/1)).
is_codelist([A|L]) :- maplist(is_codelist_code,[A|L]).
is_codelist_code(H) :- integer(H). %, swish_render_codes_charset_code(ascii, H),
:- export(is_codelist/1).
:- endif.

:- if( \+ current_predicate(is_charlist/1)).
is_charlist([A|L]) :- maplist(is_charlist_char,[A|L]).
is_charlist_char(H) :- atom(H), atom_length(H,1). 
:- export(is_charlist/1).
:- endif.

%% tokenize(+ACEText:term, -Tokens:list) is det.
%
% Breaks the ACEText (either an atom, or a list of character codes) into
% a list of tokens (atoms or numbers).
% The input ACEText can either be an atom like 'this is an example' or a list of character
% codes like [116,104,105,115,32,105,...] (possibly written as "this is an example").
%
% Tokenization will never fail. In case something goes wrong (e.g. a string
% or comment is not closed) then an error message is asserted.
%
% @param ACEText is the input text, it is either an atom or a string
% @param Tokens is a list of tokens, i.e. the tokenization of the input text
%
tokenize([], []) :- !.

tokenize(Codes, Tokens) :-  is_codelist(Codes),!,
   codes_to_tokens(Codes, Tokens),!.
tokenize(Chars, Tokens) :- (is_charlist(Chars);atom(Chars);string(Chars);number(Chars)), 
   string_codes(Chars,Codes),!,codes_to_tokens(Codes, Tokens).
tokenize(S, Tokens) :- any_to_string(S,String), string_codes(String,Codes),!,codes_to_tokens(Codes, Tokens).

tokenize_NEW([], []) :- !.

tokenize_NEW(Text, Tokens) :-
	(   atom(Text)
	->  atom_codes(Text, Codes)
	;   string(Text)
	->  string_codes(Text, Codes)
	),
	!,
	codes_to_tokens(Codes, Tokens).

tokenize_NEW(Codes, Tokens) :-
	must_be(codes, Codes),
	codes_to_tokens(Codes, Tokens).


%% codes_to_tokens(+Codes:list, -Tokens:list) is det.
%
% Maps the given list of character codes into a list of tokens.
% Performs two steps:
%
% 1. Merges certain code sequences into single tokens (atoms or numbers).
% 2. Modifies certain token sequences, e.g. [can, ', t] -> [can, not].
%
% @param Codes is a list of character codes
% @param Tokens is a list of tokens, i.e. the tokenization of the input text
%
codes_to_tokens(Codes, Tokens) :-
	get_atomics(Codes, Atomics),
	expand_contracted_forms0(Atomics, Tokens).


%% get_atomics(+Codes:list, -Tokens:list) is det.
%
% BUG: We could also preserve the comment in the token list
% to be able to show better error messages.
%
get_atomics([], []).

% A paragraph break starts with \n followed by whitespace,
% including at least one more \n.
% If another character is encountered first then the paragraph is not broken.
get_atomics([10 | Cs], AllTokens) :-
	get_whitespace_and_comments(Cs, Remaining, Newline_Count),
	!,
	add_paragraph_break_symbol(Newline_Count, Ts, AllTokens),
	get_atomics(Remaining, Ts).

% Quoted string starts with " (34)
get_atomics([34 | Cs], AllTokens) :-
	!,
	get_string(Cs, Prefix, Remaining),
	string_to_token(Prefix, Ts, AllTokens),
	get_atomics(Remaining, Ts).

% Quoted word starts with ` (96)
get_atomics([96 | Cs], [T | Ts]) :-
	!,
	get_qword(Cs, Prefix, Remaining),
	atom_codes(T, Prefix),
	get_atomics(Remaining, Ts).

% Perl-style comment starts with # (35)
get_atomics([35 | Cs], Ts) :-
	!,
	get_perl_comment(Cs, _Prefix, Remaining),
	get_atomics(Remaining, Ts).

% C-style comment starts with /* (47, 42)
get_atomics([47, 42 | Cs], Ts) :-
	!,
	get_c_comment(Cs, _Prefix, Remaining),
	get_atomics(Remaining, Ts).

% Whitespace (\t, ' ', \r) is ignored (whitespace does not include \n)
get_atomics([C | Cs], Ts) :-
	is_whitespace(C),
	!,
	get_atomics(Cs, Ts).

% Positive number starts with a digit
get_atomics([C | Cs], [Number | Ts]) :-
	is_digit(C),
	!,
	get_number(Cs, Prefix, Remaining),
	number_codes(Number, [C | Prefix]),
	get_atomics(Remaining, Ts).

% Negative number starts with a hyphen (45) and then a digit,
% e.g. numbers like `-.5' are not allowed.
get_atomics([45, C | Cs], ['-', Number | Ts]) :-
	is_digit(C),
	!,
	get_number(Cs, Prefix, Remaining),
	number_codes(Number, [C | Prefix]),
	get_atomics(Remaining, Ts).

% Word starts (see is_word_char/1)
get_atomics([C | Cs], AllTokens) :-
	is_word_char(C),
	!,
	get_word(Cs, Prefix, Remaining),
	split_token([C | Prefix], Ts, AllTokens),
	get_atomics(Remaining, Ts).

% Special character maps to a one-character token
get_atomics([C | Cs], [T | Ts]) :-
	is_special(C, T),
	!,
	get_atomics(Cs, Ts).

% All other characters (e.g. Japanese) are ignored, with a warning message.
get_atomics([C | Cs], Ts) :-
	with_output_to(atom(CharCode), format("~c (0x~16r, ~10r)", [C, C, C])),
	add_warning_message_once(character, '', CharCode, 'Unknown character(s) ignored.'),
	get_atomics(Cs, Ts).


%% split_token(+Codes:list, +Ts:list, -FinalTokens:list) is det.
%
% Builds a token (atom) and splits it into two if needed,
% e.g. 'Everything' is split into 'Every' and '-thing'.
%
split_token(Codes, Ts, FinalTokens) :-
	atom_codes(A, Codes),
	(
		pronoun_split(A, (A1, A2))
	->
		FinalTokens = [A1, A2 | Ts]
	;
		FinalTokens = [A | Ts]
	).


%% is_whitespace(?Code)
%
% Note that newlines are handled elsewhere.
%
is_whitespace(32). % ' '
is_whitespace(9).  % \t
is_whitespace(13). % \r


%% is_word_char(?Code)
%
% Characters that are allowed in ACE words.
%
% - (hyphen)
is_word_char(45).

% _ (underscore)
is_word_char(95).

% $ (dollar)
is_word_char(36).

% degree sign
is_word_char(176).

% letters
is_word_char(Code) :-
	is_letter(Code).


%% is_special(?Code, ?Atom) is det.
%
% Characters that translate into one-character atoms.
%
% @param Code is character code
% @param Atom is a one-character atom that corresponds to the code

% period, question mark, exclamation mark
is_special(46, '.').
is_special(63, '?').
is_special(33, '!').

% hyphen
is_special(45, '-').

% colon (for prefixed words' support)
is_special(58, ':').

% apostroph (for Saxon genitive support)
is_special(39, '\'').

% slash (for him/her support)
is_special(47, '/').

% comma (for comma-and, comma-or support)
is_special(44, ',').

% plus sign
is_special(43, '+').

% exponentiation sign
is_special(94, '^').

% star
is_special(42, '*').

% parentheses ()
is_special(40, '(').
is_special(41, ')').

% square brackets []
is_special(91, '[').
is_special(93, ']').

% curly bracktes {}
is_special(123, '{').
is_special(125, '}').

% < = > \
is_special(60, '<').
is_special(61, '=').
is_special(62, '>').
is_special(92, '\\').

% ampersand
is_special(38, '&').


%% get_string(+Codes:list, -Prefix:list, -RemainingCodes:list) is det.
%
% Consumes the sequence of characters within a quoted string,
% as well as the closing quotation mark.
% The backslash (\, 92) can be used to escape the following character,
% e.g. the quotation mark or a backslash.
%
get_string([], [], []) :-
	add_werror_message_once(character, '', 'EOF', 'Every quoted string must end with ".').

get_string([34 | Cs], [], Cs) :- !.

get_string([92, C | Cs], [C | Prefix], Remaining) :-
	!,
	get_string(Cs, Prefix, Remaining).

get_string([C | Cs], [C | Prefix], Remaining) :-
	get_string(Cs, Prefix, Remaining).


%% get_qword(+Codes:list, -Prefix:list, -RemainingCodes:list) is det.
%
% Consumes the sequence of characters within a quoted word,
% as well as the closing backtick.
% The backslash (\, 92) can be used to escape the following character,
% e.g. the quotation mark or a backslash.
%
get_qword([], [], []) :-
	add_werror_message_once(character, '', 'EOF', 'Every quoted word must end with `.').

get_qword([96 | Cs], [], Cs) :- !.

get_qword([92, C | Cs], [C | Prefix], Remaining) :-
	!,
	get_qword(Cs, Prefix, Remaining).

get_qword([C | Cs], [C | Prefix], Remaining) :-
	get_qword(Cs, Prefix, Remaining).


%% get_word(+Codes:list, -Prefix:list, -RemainingCodes:list) is det.
%
% Consumes the word consisting of word characters
% and/or digits.
%
get_word([], [], []).

get_word([C | Cs], [C | Prefix], Remaining) :-
	is_word_char(C),
	!,
	get_word(Cs, Prefix, Remaining).

get_word([C | Cs], [C | Prefix], Remaining) :-
	is_digit(C),
	!,
	get_word(Cs, Prefix, Remaining).

get_word([C | Cs], [], [C | Cs]).


%% get_number(+Codes:list, -Prefix:list, -RemainingCodes:list) is det.
%
% Consumes a number which is a sequence of digits containing
% at most one dot (46).
% The dot (if present) must be followed by a digit.
%
get_number(Codes, Prefix, RemainingCodes) :-
	get_number_x(Codes, zero, Prefix, RemainingCodes).


get_number_x([], _, [], []).

get_number_x([46, C | Cs], zero, [46, C | Prefix], Remaining) :-
	is_digit(C),
	!,
	get_number_x(Cs, more_than_zero, Prefix, Remaining).

get_number_x([C | Cs], Dot_Count, [C | Prefix], Remaining) :-
	is_digit(C),
	!,
	get_number_x(Cs, Dot_Count, Prefix, Remaining).

get_number_x([C | Cs], _, [], [C | Cs]).


%% get_perl_comment(+Codes:list, -Prefix:list, -RemainingCodes:list) is det.
%
% Consumes the Perl-style comment excluding the final newline.
%
get_perl_comment([], [], []) :-
	add_error_message_once(character, '', 'EOF', 'Every #-comment must end with the newline.').

get_perl_comment([10 | Cs], [], [10 | Cs]) :-
	!.

get_perl_comment([C | Cs], [C | Prefix], Remaining) :-
	get_perl_comment(Cs, Prefix, Remaining).


%% get_c_comment(+Codes:list, -Prefix:list, -RemainingCodes:list) is det.
%
% Consumes the C-style comment including the final */.
%
get_c_comment([], [], []) :-
	add_error_message_once(character, '', 'EOF', 'Every /*-comment must end with */.').

get_c_comment([42, 47 | Cs], [], Cs) :-
	!.

get_c_comment([C | Cs], [C | Prefix], Remaining) :-
	get_c_comment(Cs, Prefix, Remaining).


%% string_to_token
%
% The content of the string is bordered by quotation marks
% in the token list.
%
% BUG: I'd rather prefer a representation such as
% string('') or string(content).
%
string_to_token(Prefix, Ts, [T | Ts]) :-
	atom_codes(S, Prefix),
	concat_atom(['"', S, '"'], T).


%% add_paragraph_break_symbol
%
add_paragraph_break_symbol(at_least_two, Ts, ['<p>' | Ts]) :- !.
add_paragraph_break_symbol(_, Ts, Ts).


%% get_whitespace_and_comments
%
% Consuming whitespace and counting the newlines.
%
get_whitespace_and_comments(Cs, Remaining, Newline_Count) :-
	get_whitespace_and_comments(Cs, one, Remaining, Newline_Count).

get_whitespace_and_comments([], Newline_Count, [], Newline_Count).

get_whitespace_and_comments([10 | Cs], _, Remaining, Newline_Count) :-
	!,
	get_whitespace_and_comments(Cs, at_least_two, Remaining, Newline_Count).

get_whitespace_and_comments([C | Cs], Newline_Count, Remaining, Final_Newline_Count) :-
	is_whitespace(C),
	!,
	get_whitespace_and_comments(Cs, Newline_Count, Remaining, Final_Newline_Count).

/*
% Perl-style comment starts
get_whitespace_and_comments([35 | Cs], Newline_Count, Final_Remaining, Final_Newline_Count) :-
	!,
	get_perl_comment(Cs, _Prefix, Remaining),
	get_whitespace_and_comments(Remaining, Newline_Count, Final_Remaining, Final_Newline_Count).

% C-style comment starts
get_whitespace_and_comments([47, 42 | Cs], Newline_Count, Final_Remaining, Final_Newline_Count) :-
	!,
	get_c_comment(Cs, _Prefix, Remaining),
	get_whitespace_and_comments(Remaining, Newline_Count, Final_Remaining, Final_Newline_Count).
*/

get_whitespace_and_comments([C | Cs], Newline_Count, [C | Cs], Newline_Count).

ensure_punct(Allowed,X,M):- \+ \+ memberchk(punct,Allowed), !, ensure_punct2(X,M).
ensure_punct(_,X,X).
ensure_punct2(X,X):- last(X,Punct),is_punct_char(Punct),!.
ensure_punct2(X,M):- append(X,['.'],M),!.

chk_is_allowed(IsAllowed,Allowed):- \+ \+ memberchk(IsAllowed,Allowed),  IsAllowed \== unsafe.

do_idiomatic_replacements(Allowed,X,O):- do_idiomatic_replaces(Allowed,[],X,Y), ensure_punct(Allowed,Y,O).
replace_once(X,Find,Replace,M):- append(Find,Right,NowFind), append(Left,NowFind,X),append([Left,Replace,Right],M),!.
replace_once([XX|XXX],Find,Replace,M):- downcase_atom(XX,XD),XD\==XX,!,replace_once([XD|XXX],Find,Replace,M).

do_idiomatic_replaces(Allowed,Except,X,Y):- idiomatic_replacement_shorter(IsAllowed,Find,Replace), chk_is_allowed(IsAllowed,Allowed),
  replace_once(X,Find,Replace,M), do_idiomatic_replaces(Allowed,Except,M,Y).
do_idiomatic_replaces(Allowed,Except,X,Y):- idiomatic_replacement(IsAllowed,Find,Replace), \+ member(f(Find),Except), chk_is_allowed(IsAllowed,Allowed),
  replace_once(X,Find,Replace,M), do_idiomatic_replaces(Allowed,[f(Find)|Except],M,Y).
do_idiomatic_replaces(Allowed,Except,X,Y):- idiomatic_replacement(IsAllowed,Find,Replace),\+ member(f2(Find),Except), chk_is_allowed(IsAllowed,Allowed),
  replace_once(X,Find,Replace,M), do_idiomatic_replaces(Allowed,[f2(Find)|Except],M,Y).

do_idiomatic_replaces(_,_,X,X).

is_punct_char(Punct):- atom(Punct), atom_length(Punct,1), char_type(Punct,punct).


have_been_do(have,pl).
have_been_do(has,sg).
have_been_do(do,pl).
have_been_do(does,sg).
have_been_do(be,_).
have_been_do(are,pl).
have_been_do(is,sg).
have_been_do(HD):- have_been_do(HD,_).

in_the_time(in,the).

idiomatic_replacement_shorter(tense,[is,is],[is]).
idiomatic_replacement_shorter(tense,[is,being],[is]).
idiomatic_replacement_shorter(tense,[is,be],[is]).
idiomatic_replacement_shorter(tense,[will,have],[has]).
idiomatic_replacement_shorter(tense,[will,not,have],[has,not]).
idiomatic_replacement_shorter(tense,[In,The,Past,In,The,Past],[In,The,Past]):- in_the_time(In,The).

idiomatic_replacement_shorter(tense,[In,The,Past,HaveDo,In,The,Past],[In,The,Past]):- have_been_do(HaveDo), in_the_time(In,The).
idiomatic_replacement_shorter(tense,[HaveDo,In,The,Past],[In,The,Past]):- have_been_do(HaveDo), in_the_time(In,The).
idiomatic_replacement_shorter(tense,[In,The,Past,HaveDo],[In,The,Past]):- have_been_do(HaveDo), in_the_time(In,The).

idiomatic_replacement_shorter(tense,[some,of,some],[some]).
idiomatic_replacement_shorter(acetoks,[-,B],[AB]):- freeze(B,atomic_list_concat([-,B],AB)).
idiomatic_replacement_shorter(acetoks,[A,:,B],[AB]):- nop(is_prefix(A)), freeze(B,atomic_list_concat([A,:,B],AB)).
idiomatic_replacement_shorter(acetoks,Atoms,[AB]):- recombine(Atom), atom_chars(Atom,Atoms),atom_chars(AB,Atoms).
idiomatic_replacement_shorter(tense,[Is,Verbing],[VerbR]):- freeze(Verbing,ing_to_active_verb(Verbing,Verb,Verbs)),
  freeze(Is,member(Is-VerbR,[is-Verbs,am-Verbs,are-Verb])).

ing_to_active_verb(Verbing,Verb,Verbs):- verb_forms(Verb,Verbs,_,Verbing,_).

recombine('/*').
recombine('*/').
%recombine("'s").
recombine('""').
recombine("''").

is_prefix(A):- functionwords:( noun_prefix(A,_); propername_prefix(A,_); verb_prefix(A); modif_prefix(A)).


idiomatic_replacement(tense,[he],[somebody,that,is,male]).
idiomatic_replacement(tense,[him],[somebody,that,is,male]).
idiomatic_replacement(tense,[his],[somebody,that,is,male,'''',s]).
idiomatic_replacement(tense,[himself],[somebody,that,is,male]).
idiomatic_replacement(tense,[she],[somebody,that,is,female]).
idiomatic_replacement(tense,[herself],[somebody,that,is,female]).
idiomatic_replacement(tense,[her],[somebody,that,is,female]).
%idiomatic_replacement(tense,[her],[somebody,that,is,female,'''',s]).
idiomatic_replacement(tense,[hers],[something,of,her]).
idiomatic_replacement(tense,[it],[something,that,is,genderless]).
idiomatic_replacement(tense,[you],[somebody,that,is,the,hearer]).
idiomatic_replacement(tense,[your],[somebody,that,is,the,hearer,'''',s]).
idiomatic_replacement(tense,[yourself],[somebody,that,is,the,hearer]).
idiomatic_replacement(tense,['i'],[somebody,that,is,the,speaker]).
%idiomatic_replacement(tense,['I'],[somebody,that,is,the,speaker]).
idiomatic_replacement(tense,['myself'],[somebody,that,is,the,speaker]).
idiomatic_replacement(tense,['me'],[somebody,that,is,the,speaker]).
idiomatic_replacement(tense,[my],[somebody,that,is,the,speaker,'''',s]).
idiomatic_replacement(tense,[we],[some,'persons-we']).
idiomatic_replacement(tense,[ourself],[some,'persons-we']).
idiomatic_replacement(tense,[ourselves],[some,'persons-we']).
idiomatic_replacement(tense,[us],[some,'persons-we']).
idiomatic_replacement(tense,[those],[some,'those-they']).
idiomatic_replacement(tense,[these],[some,'those-they']).
idiomatic_replacement(tense,[their],['those-they','''',s]).
idiomatic_replacement(tense,[they],[some,'those-they']).
idiomatic_replacement(tense,[them],[some,'those-they']).
%idiomatic_replacement(unsafe,[does,some],[is,doing]).
%idiomatic_replacement(unsafe,[does],[is,doing]).
idiomatic_replacement(tense,[wont],[will,not]).
idiomatic_replacement(tense,['"',Eat],['"',Eats]):- freeze(Eat,verb_forms(Eat,Eats,_Ate,_Eating,_Eaten)).
%idiomatic_replacement(tense,[was],[that,is,In,The,past,that,is]).
%idiomatic_replacement(tense,[will],[that,is,In,The,future,that]).

%idiomatic_replacement(C,F,R):- idiomatic_replacement(C,F,R2,R1),append(R1,R2,R).
/*
idiomatic_replacement_prepends(tense,Before,After,[if,The,statement,is,Old,statement,then]):-
  idiomatic_replacement_prepends(tense,Before,After,[In,The,Past]),
   (Past == past -> Old = old ; Old = new).
*/

idiomatic_replacement(tense,[was],[In,The,past,is]):- in_the_time(In,The).
idiomatic_replacement(tense,[were],[In,The,past,are]):- in_the_time(In,The).

idiomatic_replacement(tense,[did],[In,The,past]):- in_the_time(In,The).
idiomatic_replacement(tense,[will],[In,The,future]):- in_the_time(In,The).

idiomatic_replacement(tense,[could],[In,The,past,can]):- in_the_time(In,The).

idiomatic_replacement(tense,[has],[In,The,past,have]):- in_the_time(In,The).
idiomatic_replacement(tense,[had],[In,The,past,have]):- in_the_time(In,The).
idiomatic_replacement(tense,[has,done],[In,The,past,have]):- in_the_time(In,The).

/*

idiomatic_replacement(tense,[Eat],[VEat]):-
  freeze(Eat,(verb_forms(Eat,_Eats,_Ate,_Eating,_Eaten),atom_concat('v:',Eat,VEat))).

idiomatic_replacement(tense,[Eats],[VEat]):-
  freeze(Eats,(verb_forms(Eat,Eats,_Ate,_Eating,_Eaten),atom_concat('v:',Eat,VEat))).

idiomatic_replacement(tense,[Eating],[VEat]):-
  freeze(Eating,(verb_forms(Eat,_Eats,_Ate, Eating,_Eaten),atom_concat('v:',Eat,VEat))).
*/

idiomatic_replacement(tense,[Eaten],[In,The,past,Eating]):- in_the_time(In,The),
  freeze(Eaten,(verb_forms(_Eat,_Eats,_Ate,Eating,Eaten))).

idiomatic_replacement(tense,[Ate],[In,The,past,Eating]):- in_the_time(In,The),
  freeze(Ate,(verb_forms(_Eat,_Eats,Ate,Eating,_Eaten))).

idiomatic_replacement(tense,[Past,Verb],[Past,Verbing]):- past_or_future(Past),
  freeze(Verb,any_verb_form(Verb,Verbing)).

past_or_future(past).
past_or_future(future).

any_verb_form(Verb,Verbing):- 
  arg(_,verb_forms(Eat,Eats,Ate,Eaten),Verb), 
        verb_forms(Eat,Eats,Ate,Verbing,Eaten).

verb_forms(Eat,Eats,Ate,Eating,Eaten):- talk_db(_,Eat,Eats,Ate,Eating,Eaten),
  \+ skip_verb(Eat), \+ skip_verb(Eats), \+ skip_verb(Ate), \+ skip_verb(Eating), \+ skip_verb(Eaten),  
  %\+ verbs:aux_modal_verb(Eat).
  true.

skip_verb(do).
skip_verb(be).
skip_verb(is).
skip_verb(time).
skip_verb(point).
skip_verb(in).
skip_verb(can).
%skip_verb(X):- clex:noun_pl(_,X,_).
%skip_verb(X):- clex:noun_sg(_,X,_).
skip_verb(have).
skip_verb(named).

%% expand_contracted_forms(+TokenListIn:list, -TokenListOut:list) is det.
%
% @bug: `cannot' could be instead handled during (pronoun) splitting
%

expand_contracted_forms(I,O):- expand_contracted_forms0(I,M),show_failure(always,do_idiomatic_replacements([acetoks],M,O)).
expand_contracted_forms(MayDo,I,O):- expand_contracted_forms0(I,M),show_failure(always,do_idiomatic_replacements(MayDo,M,O)).

expand_contracted_forms0([], []).

expand_contracted_forms0(['No', one | RestIn], ['No', '-one' | RestOut]) :-
	!,
	expand_contracted_forms0(RestIn, RestOut).

expand_contracted_forms0([no, one | RestIn], [no, '-one' | RestOut]) :-
	!,
	expand_contracted_forms0(RestIn, RestOut).

expand_contracted_forms0([isn, '\'', t | RestIn], [is, not | RestOut]) :-
	!,
	expand_contracted_forms0(RestIn, RestOut).

expand_contracted_forms0([aren, '\'', t | RestIn], [are, not | RestOut]) :-
	!,
	expand_contracted_forms0(RestIn, RestOut).

expand_contracted_forms0([doesn, '\'', t | RestIn], [does, not | RestOut]) :-
	!,
	expand_contracted_forms0(RestIn, RestOut).

expand_contracted_forms0([don, '\'', t | RestIn], [do, not | RestOut]) :-
	!,
	expand_contracted_forms0(RestIn, RestOut).

expand_contracted_forms0([can, '\'', t | RestIn], [can, not | RestOut]) :-
	!,
	expand_contracted_forms0(RestIn, RestOut).

expand_contracted_forms0([cannot | RestIn], [can, not | RestOut]) :-
	!,
	expand_contracted_forms0(RestIn, RestOut).

expand_contracted_forms0([shouldn, '\'', t | RestIn], [should, not | RestOut]) :-
	!,
	expand_contracted_forms0(RestIn, RestOut).
/*
expand_contracted_forms0(TextIn, AfterRestOut) :- freeze(A,(A==contractions;A==slang;A==mispellings)),dictionary(A,Before,After),
  listify(Before,BeforeL),append(BeforeL,RestIn,TextIn), !,
	expand_contracted_forms0(RestIn, RestOut),
  listify(After,AfterL),append(AfterL,RestOut,AfterRestOut).
*/
expand_contracted_forms0([Token | TailIn], [Token | TailOut]) :-
	expand_contracted_forms0(TailIn, TailOut).



